from . import types
from . import base
from . import core
from . import cloud
from . import generic

from .generic import as_path, get_path, PathLike, register_pathlike_cls, get_userhome, get_lazydir, Union, Type

to_path = get_path

from .core import PosixFSxPath, WindowsFSxPath
from .cloud import PosixGCSPath, PosixS3Path, WindowsS3Path, WindowsGCSPath

import pathlib

PathzPath = (PosixFSxPath, WindowsFSxPath, PosixGCSPath, PosixS3Path, WindowsS3Path, WindowsGCSPath, pathlib.Path, pathlib.PosixPath, pathlib.PurePath, str)
PathzLike = Union[Type[PosixFSxPath], Type[WindowsFSxPath], Type[PosixGCSPath], Type[PosixS3Path], Type[WindowsS3Path], Type[WindowsGCSPath], Type[pathlib.PosixPath], Type[pathlib.PurePath], Type[core.ReadOnlyPath], Type[core.ReadWritePath]]


from . import reqz

from .reqz import *

""" 
Borrowed from

https://github.com/slingamn/mureq

Why?
In short: performance (memory consumption), security (resilience to supply-chain attacks), and simplicity.

Performance
python-requests is extremely memory-hungry, mainly due to large transitive dependencies like chardet that are not needed

mureq is a replacement for python-requests, intended to be vendored
in-tree by Linux systems software and other lightweight applications.

mureq is copyright 2021 by its contributors and is released under the
0BSD ("zero-clause BSD") license.

---
You can either use httpz or reqz which are identical

from lazy import httpz
from lazy.httpz import reqz

print(httpz.get('https://clients3.google.com/generate_204'))
print(reqz.get('https://clients3.google.com/generate_204'))

"""


"""
CLI to work with lazy.serialize

- Tested to work with local, s3, gs

Useful examples:

- working with secrets like AWS_ACCESS_KEY_ID, AWS_SECRET_ACCESS_KEY
---
echo "$AWS_ACCESS_KEY_ID" | xargs lazy encode b64
>>> QUtJ... 
echo "$AWS_ACCESS_KEY_ID" | xargs lazy encode bgz
>>> H4s...

echo "$AWS_ACCESS_KEY_ID" | export AWS_KEY_ID_BGZ=$(xargs lazy encode bgz)
echo "$AWS_KEY_ID_BGZ" | xargs lazy decode bgz
>>> A...
---

- working with service_account credentials
--- 
lazy encode jsonbgz --input-path '/path/to/adc.json'
>>> H4sIAJxJ52EC/5VWWa...

lazy decode jsonbgz $ADC_BGZ --output-path '/path/to/adc.json'
>>> /path/to/adc.json
---

- read & replace a template file
--- 
lazy read replace 's3://bucket/to/config_template.yaml' --output-path '/local/path/config.yaml' \
    --fix "<CLUSTER_NAME>=$CLUSTER_NAME" \
    --fix "<CLUSTER_VER>=$CLUSTER_VER" \
    --fix "<IMAGE_TAG>=$IMAGE_TAG"
>>> /local/path/config.yaml
---
"""

from .base import *
from lazy.serialize import Serializer

__all__ = ( 
    'encodeCli',
    'decodeCli',
    'keysCli',
    'readCli',
    'writeCli',
)
encodeCli = createCli(name = 'encode')
decodeCli = createCli(name = 'decode')
keysCli = createCli(name = 'keys')

readCli = createCli(name = 'read')
writeCli = createCli(name = 'write')


## Will import and run func dynamically
def _get_path(path: str):
    from lazy.io import get_path
    return get_path(path)


@readCli.command('file', short_help = "Reads from a File, writes to the --output-path [optional]")
def read_file(
        input_path: str = Argument(None),
        output_path: Optional[str] = Argument(None),
        binary: Optional[bool] = Option(False),
    ):
    input_path = _get_path(input_path)
    data = input_path.read_text() if not binary else input_path.read_binary()
    if output_path:
        output_path = _get_path(output_path)
        if binary: output_path.write_binary(data)
        else: output_path.write_text(data)
        echo(output_path.as_posix())
    else:
        echo(data)


@readCli.command('replace', short_help = "Reads from a File, replaces --fix x=y=num [sed-like], writes to the --output-path [optional]")
def read_replace_file(
        input_path: str = Argument(None),
        fix: Optional[List[str]] = Option(None),
        output_path: Optional[str] = Option(None),
        binary: Optional[bool] = Option(False),
    ):
    input_path = _get_path(input_path)
    data = input_path.read_text() if not binary else input_path.read_binary()
    for f in fix:
        change = f.split('=')
        data = data.replace(str(change[0]), str(change[1]), int(change[2])) if len(change) == 3 else data.replace(str(change[0]), str(change[1]))
    if output_path:
        output_path = _get_path(output_path)
        if binary: output_path.write_binary(data)
        else: output_path.write_text(data)
        echo(output_path.as_posix())
    else:
        echo(data)


@writeCli.command('file', short_help = "Write to a File from text or from --input-path [optional], writes it to the --output-path [optional]")
def write_file(
        text: Optional[str] = Argument(None),
        input_path: str = Option(None),
        output_path: Optional[str] = Option(None),
        append: Optional[bool] = Option(False),
        binary: Optional[bool] = Option(False),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = input_path.read_text() if not binary else input_path.read_binary()
    else:
        data = text.encode('utf-8') if binary else text
    if output_path:
        output_path = _get_path(output_path)
        if binary: output_path.write_binary(data)
        else: 
            mode = 'a' if append else 'w'
            with output_path.open(mode = mode) as f:
                f.write(data)
                if append: f.write('\n')
        echo(output_path.as_posix())
    else:
        echo(data)


@decodeCli.command('json', short_help = "Decodes JSON string or JSON File into text")
def decode_json(
        text: Optional[str] = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = input_path.read_json()
    else:
        data = Serializer.DefaultJson.loads(text)
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(Serializer.DefaultJson.dumps(data))
        echo(output_path.as_posix())
    else:
        echo(Serializer.DefaultJson.dumps(data))


@encodeCli.command('jsonb64', short_help = "Encodes a JSON String or JSON File into Base64")
def encode_json_b64(
        text: Optional[str] = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = Serializer.Base.b64_encode(Serializer.DefaultJson.dumps(input_path.read_json()))
    else:
        # We do this to ensure valid JSON encoding
        data = Serializer.Base.b64_encode(Serializer.DefaultJson.dumps(Serializer.DefaultJson.loads(text)))
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(data)
        echo(output_path.as_posix())
    else:
        echo(data)


@encodeCli.command('jsonbgz', short_help = "Encodes a JSON String or JSON File into Base64 + GZIP")
def encode_json_bgz(
        text: Optional[str] = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = Serializer.Base.b64_gzip_encode(Serializer.DefaultJson.dumps(input_path.read_json()))
    else:
        # We do this to ensure valid JSON encoding
        data = Serializer.Base.b64_gzip_encode(Serializer.DefaultJson.dumps(Serializer.DefaultJson.loads(text)))
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(data)
        echo(output_path.as_posix())
    else:
        echo(data)


@decodeCli.command('jsonb64', short_help = "Decodes a Base64 String into JSON")
def decode_json_b64(
        text: Optional[str] = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = Serializer.DefaultJson.loads(Serializer.Base.b64_decode(input_path.read_text()))
    else:
        data = Serializer.DefaultJson.loads(Serializer.Base.b64_decode(text))
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(Serializer.DefaultJson.dumps(data))
        echo(output_path.as_posix())
    else:
        echo(Serializer.DefaultJson.dumps(data))


@decodeCli.command('jsonbgz', short_help = "Decodes a Base64 + GZIP String into JSON")
def decode_json_bgz(
        text: Optional[str] = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = Serializer.DefaultJson.loads(Serializer.Base.b64_gzip_decode(input_path.read_text()))
    else:
        data = Serializer.DefaultJson.loads(Serializer.Base.b64_gzip_decode(text))
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(Serializer.DefaultJson.dumps(data))
        echo(output_path.as_posix())
    else:
        echo(Serializer.DefaultJson.dumps(data))


@decodeCli.command('pickle')
def decode_pickle(
        text: str = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = Serializer.Pkl.loads(input_path.read_bytes())
    else:
        data = Serializer.Pkl.loads(text.encode('utf-8'))
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_bytes(Serializer.Pkl.dumps(data))
        echo(output_path.as_posix())
    else:
        echo(data)


@encodeCli.command('pickle')
def encode_pickle(
        text: str = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = input_path.read_bytes()
    else:
        data = Serializer.Pkl.dumps(text.encode('utf-8'))
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_bytes(data)
        echo(output_path.as_posix())
    else:
        echo(data)


@decodeCli.command('b64', short_help = "Decodes string or Text File from Base64")
def decode_b64(
        text: Optional[str] = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = Serializer.Base.b64_decode(input_path.read_text())
    else:
        data = Serializer.Base.b64_decode(text)
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(data)
        echo(output_path.as_posix())
    else:
        echo(data)


@encodeCli.command('b64', short_help = "Encodes string or Text File to Base64")
def encode_b64(
        text: Optional[str] = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = Serializer.Base.b64_encode(input_path.read_text())
    else:
        data = Serializer.Base.b64_encode(text)
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(data)
        echo(output_path.as_posix())
    else:
        echo(data)


@decodeCli.command('bgz', short_help = "Decodes string or Text File from Base64 + GZIP")
def decode_bgz(
        text: Optional[str] = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = Serializer.Base.b64_gzip_decode(input_path.read_text())
    else:
        data = Serializer.Base.b64_gzip_decode(text)
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(data)
        echo(output_path.as_posix())
    else:
        echo(data)


@encodeCli.command('bgz', short_help = "Encodes string or Text File to Base64 + GZIP")
def encode_bgz(
        text: Optional[str] = Argument(None),
        input_path: Optional[str] = Option(None),
        output_path: Optional[str] = Option(None),
    ):
    if input_path:
        input_path = _get_path(input_path)
        data = Serializer.Base.b64_gzip_encode(input_path.read_text())
    else:
        data = Serializer.Base.b64_gzip_encode(text)
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(data)
        echo(output_path.as_posix())
    else:
        echo(data)


@keysCli.command('uuid', short_help = "Creates a UUID from UUID4")
def create_uuid(
        length: Optional[int] = Argument(None), 
        output_path: Optional[str] = Option(None),
        clean: bool = Option(False),
    ):
    result_uid = ''
    if length:
        while len(result_uid) <= length:
            _uid = Serializer.Base.get_uuid()
            if clean: _uid = _uid.replace('-', '')
            result_uid += _uid
        result_uid = result_uid[:length]
    else:
        result_uid = Serializer.Base.get_uuid()
        if clean: result_uid = result_uid.replace('-', '')
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(result_uid)
        echo(output_path.as_posix())
    else:
        echo(result_uid)


@keysCli.command('secret', short_help = "Creates an Alphanumeric Secret Passcode")
def create_alpha_passcode(
        length: Optional[int] = Argument(32),
        output_path: Optional[str] = Option(None),
    ):
    result = Serializer.Secret.alphanumeric_passcode(length=length)
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(result)
        echo(output_path.as_posix())
    else:
        echo(result)


@keysCli.command('keypair', short_help = "Creates a Keypair [access_key | secret_key]")
def create_keypair(
        key_length: Optional[int] = Option(16), 
        secret_length: Optional[int] = Option(32),
        output_path: Optional[str] = Option(None),
    ):
    result = Serializer.Secret.keypair(k_len=key_length, s_len=secret_length)
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(result)
        echo(output_path.as_posix())
    else:
        echo(result)


@keysCli.command('token', short_help = "Creates a Token [API Token]")
def create_token(
        length: Optional[int] = Argument(16),
        clean: Optional[bool] = Option(True), 
        safe: Optional[bool] = Option(False),
        output_path: Optional[str] = Option(None),
    ):
    result = Serializer.Secret.token(length=length, safe=safe, clean=clean)
    if output_path:
        output_path = _get_path(output_path)
        output_path.write_text(result)
        echo(output_path.as_posix())
    else:
        echo(result)




import os
import pathlib
from .base import *
from .providers.gs_gcp import *
from .providers.s3_aws import *
from .providers.s3_minio import *

from typing import List, Dict, Union, Type, Tuple

PathzLike = Union[
    Type[PathzPurePath],
    Type[PathzPath],
    Type[PurePathzPosixPath],
    Type[PathzWindowsPath],
    Type[PathzPosixPath],
    Type[PurePathzWindowsPath],
    Type[PathzGSPurePath],
    Type[PathzGSPath],
    Type[PurePathzGSPosixPath],
    Type[PathzGSWindowsPath],
    Type[PathzGSPosixPath],
    Type[PurePathzGSWindowsPath],
    Type[PathzS3PurePath],
    Type[PathzS3Path],
    Type[PurePathzS3PosixPath],
    Type[PathzS3WindowsPath],
    Type[PathzS3PosixPath],
    Type[PurePathzS3WindowsPath],
    Type[PathzMinioPurePath], Type[PathzMinioPath], Type[PurePathzMinioPosixPath], Type[PathzMinioWindowsPath], Type[PathzMinioPosixPath], Type[PurePathzMinioWindowsPath]

]

_PATHLIKE_CLS: Tuple[PathzLike, ...] = (
    PathzPurePath,
    PathzPath,
    PurePathzPosixPath,
    PathzWindowsPath,
    PathzPosixPath,
    PurePathzWindowsPath,
    PathzGSPurePath,
    PathzGSPath,
    PurePathzGSPosixPath,
    PathzGSWindowsPath,
    PathzGSPosixPath,
    PurePathzGSWindowsPath,
    PathzS3PurePath,
    PathzS3Path,
    PurePathzS3PosixPath,
    PathzS3WindowsPath,
    PathzS3PosixPath,
    PurePathzS3WindowsPath,
    PathzMinioPurePath, PathzMinioPath, PurePathzMinioPosixPath, PathzMinioWindowsPath, PathzMinioPosixPath, PurePathzMinioWindowsPath
)

FileSysLike = Union[
    Type[AWSFileSystem],
    Type[GCPFileSystem],
    Type[MinioFileSystem],
]

PathLike = Union[str, os.PathLike, PathzLike]

_PREFIXES_TO_CLS: Dict[str, PathzLike] = {
    'gs://': PathzGSPath,
    's3://': PathzS3Path,
    'minio://': PathzMinioPath,
    #'minio://': cloud.PosixMinioPath,
    #'s3compat://': cloud.PosixS3CompatPath,
}


def as_path(path: PathLike) -> PathzLike:
    """Create a generic `pathlib.Path`-like abstraction.
    Depending on the input (e.g. `gs://`, `github://`, `ResourcePath`,...), the
    system (Windows, Linux,...), the function will create the right pathlib-like
    abstraction.
    Args:
        path: Pathlike object.
    Returns:
        path: The `pathlib.Path`-like abstraction.
    """
    if isinstance(path, str):
        uri_splits = path.split('://', maxsplit=1)
        if len(uri_splits) > 1:    
            # str is URI (e.g. `gs://`, `github://`,...)
            return _PREFIXES_TO_CLS[uri_splits[0] + '://'](path)
        return PathzPath(path)
    elif isinstance(path, _PATHLIKE_CLS):
        return path
    elif isinstance(path, os.PathLike):
        return PathzPath(path)
    else: raise TypeError(f'Invalid path type: {path!r}')


def get_userhome(as_pathz: bool = True):
    h = os.path.expanduser('~')
    if as_pathz: return as_path(h)
    return h

def get_cwd():
    return os.getcwd()


def resolve_relative(filepath: PathLike) -> str:
    if not isinstance(filepath, str): filepath = filepath.as_posix()
    if '://' in filepath: return filepath
    if filepath.startswith('~'): filepath = filepath.replace('~', get_userhome(), 1)
    elif filepath.startswith('../'): filepath = filepath.replace('..', get_cwd(), 1)
    elif filepath.startswith('..'): filepath = filepath.replace('..', pathlib.Path(get_cwd()).parent.parent.as_posix() + '/', 1)
    elif filepath.startswith('./'): filepath = filepath.replace('.', get_cwd(), 1)
    elif filepath.startswith('.'): filepath = filepath.replace('.', pathlib.Path(get_cwd()).parent.as_posix() + '/', 1)
    return filepath

def get_path(filepath: PathLike, resolve: bool = False) -> PathzLike:
    if resolve: filepath = resolve_relative(filepath)
    if isinstance(filepath, str): filepath = as_path(filepath)
    return filepath

def get_pathlike(filepath: PathLike, resolve: bool = False) -> PathzLike:
    if resolve: filepath = resolve_relative(filepath)
    if isinstance(filepath, str): filepath = as_path(filepath)
    return filepath

def get_lazydir(mkdir: bool = False):
    """
    returns resolved '~/.lazy'
    """
    p = get_userhome()
    p = p.joinpath('.lazy')
    if mkdir: p.mkdir(exist_ok=True, parents=True)
    return p

to_path = get_path
__all__ = (
    'get_path', 'as_path', 'PathLike', 'PathzPath', 'PathzS3Path', 'PathzGSPath', 'PathzLike', 'get_lazydir',
)


from __future__ import annotations
import os
from .base import *


from .cfs_base import GCP_CFS
from .cfs_pathz_base import *


class PathzGSPurePath(PathzCFSPurePath):
    _prefix: str = 'gs'
    _provider: str = 'GoogleCloudStorage'
    _win_pathz: ClassVar = 'PurePathzGSWindowsPath'
    _posix_pathz: ClassVar = 'PurePathzGSPosixPath'


class PurePathzGSPosixPath(PurePathzCFSPosixPath):
    """PurePath subclass for non-Windows systems.
    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _async_sync_posix_flavour
    _pathlike = posixpath
    __slots__ = ()


class PurePathzGSWindowsPath(PurePathzCFSWindowsPath):
    """PurePath subclass for Windows systems.
    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _async_sync_windows_flavour
    _pathlike = ntpath
    __slots__ = ()

class PathzGSPath(PathzCFSPath):
    """
    Our customized class that incorporates both sync and async methods
    """
    _flavour = _async_sync_windows_flavour if os.name == 'nt' else _async_sync_posix_flavour
    _accessor: AccessorLike = None
    _pathlike = posixpath
    _prefix = 'gs'
    _provider = 'GoogleCloudStorage'

    _win_pathz: ModuleType = 'PathzGSWindowsPath'
    _posix_pathz: ModuleType = 'PathzGSPosixPath'

    def _init(self, template: Optional['PathzGSPath'] = None):
        self._accessor: AccessorLike = get_accessor(self._prefix)
        self._closed = False
        self._fileio = None

    def __new__(cls, *parts, **kwargs):
        if cls is PathzGSPath or issubclass(PathzGSPath): 
            cls = cls._win_pathz if os.name == 'nt' else cls._posix_pathz
            cls = globals()[cls]
        self = cls._from_parts(parts, init=False)
        if not self._flavour.is_supported:
            name: str = cls.__name__
            raise NotImplementedError(f"cannot instantiate {name} on your system")

        self._init()
        return self


class PathzGSPosixPath(PosixPath, PathzGSPath, PurePathzGSPosixPath):
    __slots__ = ()


class PathzGSWindowsPath(WindowsPath, PathzGSPath, PurePathzGSWindowsPath):
    __slots__ = ()

    def is_mount(self) -> int:
        raise NotImplementedError("PathzGSPath.is_mount() is unsupported on this system")

    async def async_is_mount(self) -> int:
        raise NotImplementedError("PathzGSPath.async_is_mount() is unsupported on this system")

register_pathlike(
    [
        PathzGSPurePath, PathzGSPath, PurePathzGSPosixPath, PathzGSWindowsPath, PathzGSPosixPath, PurePathzGSWindowsPath
    ]
)

GCPFileSystem = GCP_CFS


__all__ = (
    'PathzGSPurePath',
    'PathzGSPath',
    'PurePathzGSPosixPath',
    'PathzGSWindowsPath',
    'PathzGSPosixPath',
    'PurePathzGSWindowsPath',
    'GCPFileSystem'
)





from .base import baseCli
from .serialize_app import *

baseCli.add_typer(encodeCli, name='encode')
baseCli.add_typer(decodeCli, name='decode')
baseCli.add_typer(keysCli, name='keys')
baseCli.add_typer(readCli, name='read')
baseCli.add_typer(writeCli, name='write')


if __name__ == '__main__':
    baseCli()


from __future__ import annotations

import os
import io
import inspect
import ntpath
import pathlib
import posixpath
from pathlib import PosixPath, WindowsPath, Path, PurePath
from pathlib import _NormalAccessor as NormalAccessor
from typing import Optional, List, Union, AsyncIterable, Iterable, IO, TYPE_CHECKING, AsyncContextManager, cast, Callable
from os import stat_result, PathLike
from contextlib import asynccontextmanager

from anyio import open_file
#from aiopath.wrap import coro_as_method_coro, func_as_method_coro, to_thread, method_as_method_coro, func_to_async_func
#from aiopath.handle import IterableAIOFile, get_handle
#from aiopath.types import Final, Literal, FileMode

from ..aiopathz.wrap import coro_as_method_coro, func_as_method_coro, to_thread, method_as_method_coro, func_to_async_func
from ..aiopathz.handle import IterableAIOFile, get_handle
from ..aiopathz.types import Final, Literal, FileMode

from fsspec.asyn import AsyncFileSystem
from anyio import AsyncFile, open_file
from lazy.serialize import Serialize
from lazy.io.pathz_v2.types import *
#from ..types import *
from .cloud_static import _ASYNC_SYNTAX_MAPPING

if TYPE_CHECKING:  # keep mypy quiet
    from lazy.io.pathz_v2.base import PathzPath

URI_PREFIXES = ('gs://', 's3://', 'minio://', 's3compat://')
_URI_SCHEMES = frozenset(('gs', 's3', 'minio', 's3compat'))
_URI_MAP_ROOT = {
    'gs://': '/gs/',
    's3://': '/s3/',
    'minio://': '/minio/',
    's3compat://': '/s3compat/'
}
_PROVIDER_MAP = {
    'gs': 'GoogleCloudStorage',
    's3': 'AmazonS3',
    'minio': 'MinIO',
    's3compat': 'S3Compatible'
}

DEFAULT_ENCODING: Final[str] = 'utf-8'
ON_ERRORS: Final[str] = 'ignore'
NEWLINE: Final[str] = '\n'

BEGINNING: Final[int] = 0
CHUNK_SIZE: Final[int] = 4 * 1_024

SEP: Final[str] = '\n'
ENCODING: Final[str] = 'utf-8'
ERRORS: Final[str] = 'replace'


Paths = Union['PathzPath', Path, str]
FileData = Union[bytes, str]


def iscoroutinefunction(obj):
    if inspect.iscoroutinefunction(obj): return True
    if hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__): return True
    return False

"""
Lol why so inconsistent.

"""

def rewrite_async_syntax(obj, provider: str = 's3'):
    """
    Basically - we're rewriting all the fsspec's async method
    from _method to async_method for syntax
    """
    _names = _ASYNC_SYNTAX_MAPPING[provider]
    for attr in dir(obj):
        if attr.startswith('_') and not attr.startswith('__'):
            attr_val = getattr(obj, attr)
            if iscoroutinefunction(attr_val) and _names.get(attr):
                setattr(obj, _names[attr], attr_val)
    return obj

Paths = Union[Path, PathLike, str]
Handle = AsyncFile


@asynccontextmanager
async def get_cloud_handle(name: Paths, mode: FileMode = 'r', buffering: int = -1, encoding: str | None = ENCODING, errors: str | None = ERRORS, newline: str | None = SEP) -> AsyncContextManager[Handle]:
    file: AsyncFile
    #if not isinstance(name, str): name = cast(IO[Union[str, bytes]], name)
    if 'b' in mode: file = await open_file(name, mode)
    else: file = await open_file(name, mode, encoding=encoding, errors=errors, newline=newline)
    yield file
    await file.aclose()


@asynccontextmanager
async def get_cloud_file(filelike: Paths) -> AsyncContextManager[Handle]:
    file: AsyncFile
    filelike = cast(IO[Union[str, bytes, os.PathLike]], filelike)
    file = AsyncFile(filelike)
    yield file
    await file.aclose()



from __future__ import annotations

import os
import io
import inspect
import ntpath
import pathlib
import posixpath
from pathlib import PosixPath, WindowsPath, _NormalAccessor, Path, PurePath, _ignore_error
from pathlib import _make_selector as _sync_make_selector
from typing import Optional, List, Union, AsyncIterable, Iterable, IO
from os import stat_result, PathLike
from stat import S_ISDIR, S_ISLNK, S_ISREG, S_ISSOCK, S_ISBLK, S_ISCHR, S_ISFIFO

from anyio import open_file

from aiopath.selectors import _make_selector
from aiopath.flavours import _async_windows_flavour, _async_posix_flavour
from aiopath.wrap import coro_as_method_coro, func_as_method_coro, to_thread, method_as_method_coro, func_to_async_func
from aiopath.handle import IterableAIOFile, get_handle
from aiopath.scandir import EntryWrapper, scandir_async, _scandir_results
from aiopath.types import Final, Literal, FileMode

from fsspec.asyn import AsyncFileSystem

from lazy.serialize import Serialize
from .types import *
from ._flavours import _async_sync_windows_flavour, _async_sync_posix_flavour
from ._cloud_static import _ASYNC_SYNTAX_MAPPING

try: import s3fs
except ImportError: s3fs: ModuleType = None

try: import gcsfs
except ImportError: gcsfs: ModuleType = None

try: from lazy import CloudAuthz
except ImportError: CloudAuthz: object = None


URI_PREFIXES = ('gs://', 's3://', 'minio://', 's3compat://')
_URI_SCHEMES = frozenset(('gs', 's3', 'minio', 's3compat'))
_URI_MAP_ROOT = {
    'gs://': '/gs/',
    's3://': '/s3/',
    'minio://': '/minio/',
    's3compat://': '/s3compat/'
}
_PROVIDER_MAP = {
    'gs': 'GoogleCloudStorage',
    's3': 'AmazonS3',
    'minio': 'MinIO',
    's3compat': 'S3Compatible'
}

DEFAULT_ENCODING: Final[str] = 'utf-8'
ON_ERRORS: Final[str] = 'ignore'
NEWLINE: Final[str] = '\n'

_s3_filesystem: 's3fs.S3FileSystem' = None
_s3_async_filesystem: 's3fs.S3FileSystem' = None

_minio_filesystem: 's3fs.S3FileSystem' = None
_minio_async_filesystem: 's3fs.S3FileSystem' = None

_s3compat_filesystem: 's3fs.S3FileSystem' = None
_s3compat_async_filesystem: 's3fs.S3FileSystem' = None

_gcs_filesystem: 'gcsfs.GCSFileSystem' = None
_gcs_async_filesystem: 'gcsfs.GCSFileSystem' = None



def iscoroutinefunction(obj):
    if inspect.iscoroutinefunction(obj): return True
    if hasattr(obj, '__call__') and inspect.iscoroutinefunction(obj.__call__): return True
    return False

"""
Lol why so inconsistent.
https://docs.google.com/spreadsheets/d/1Lj8WXPUuw8T_VWJBUaaa3U41akuBWNDbqL09hviI8oc/edit#gid=0

"""

def _rewrite_async_syntax(obj, provider: str = 's3'):
    """
    Basically - we're rewriting all the fsspec's async method
    from _method to async_method for syntax
    """
    _names = _ASYNC_SYNTAX_MAPPING[provider]
    for attr in dir(obj):
        if attr.startswith('_') and not attr.startswith('__'):
            attr_val = getattr(obj, attr)
            if iscoroutinefunction(attr_val) and _names.get(attr):
                setattr(obj, _names[attr], attr_val)
    return obj



if CloudAuthz:
    authz = CloudAuthz()
    if s3fs is not None:
        ## Setup for AWS S3 ##
        _config = {}
        if authz.aws_access_key_id:
            _config['key'] = authz.aws_access_key_id
            _config['secret'] = authz.aws_secret_access_key
        elif authz.aws_access_token: _config['token'] = authz.aws_access_token
        elif not authz.boto_config: _config['anon'] = True
        if authz.s3_config: _config['config_kwargs'] = authz.s3_config

        _s3_filesystem: 's3fs.S3FileSystem' = s3fs.S3FileSystem(**_config)
        _s3_async_filesystem: 's3fs.S3FileSystem' = _rewrite_async_syntax(s3fs.S3FileSystem(asynchronous=True, **_config))

        ## Setup for Minio S3 ##
        _config = {}
        if authz.minio_secret_key:
            _config['key'] = authz.minio_access_key
            _config['secret'] = authz.minio_secret_key
        elif authz.minio_access_token: _config['token'] = authz.minio_access_token
        _config['client_kwargs'] = {'endpoint_url': authz.minio_endpoint}
        if authz.minio_config: _config['config_kwargs'] = authz.minio_config
        
        _minio_filesystem: 's3fs.S3FileSystem'  = s3fs.S3FileSystem(**_config)
        _minio_async_filesystem: 's3fs.S3FileSystem'  = _rewrite_async_syntax(s3fs.S3FileSystem(asynchronous=True, **_config))

        ## Setup for S3 Compat ##
        _config = {}
        if authz.s3compat_secret_key:
            _config['key'] = authz.s3compat_access_key
            _config['secret'] = authz.s3compat_secret_key
        elif authz.s3compat_access_token: _config['token'] = authz.s3compat_access_token
        _config['client_kwargs'] = {'endpoint_url': authz.s3compat_endpoint}
        if authz.s3compat_region: _config['client_kwargs']['region_name'] = authz.s3compat_region
        if authz.s3compat_config: _config['config_kwargs'] = authz.s3compat_config

        _s3compat_filesystem: 's3fs.S3FileSystem' = s3fs.S3FileSystem(**_config)
        _s3compat_async_filesystem: 's3fs.S3FileSystem' = _rewrite_async_syntax(s3fs.S3FileSystem(asynchronous=True, **_config))

    if gcsfs is not None:
        _config = {}
        if authz.gcp_auth or authz.gauth or authz.gcp_authb64 or authz.gcp_authbgz: _config['token'] = authz.gcp_auth or authz.gauth or authz.gcp_authb64 or authz.gcp_authbgz
        if authz.gcloud_project or authz.google_cloud_project: _config['project'] = authz.gcloud_project or authz.google_cloud_project
        if authz.gcs_client_config: _config['client_kwargs'] = authz.gcs_client_config
        if authz.gcs_config: _config['config_kwargs'] = authz.gcs_config

        _gcs_filesystem: 'gcsfs.GCSFileSystem' = gcsfs.GCSFileSystem( **_config)
        _gcs_async_filesystem: 'gcsfs.GCSFileSystem' = _rewrite_async_syntax(gcsfs.GCSFileSystem(asynchronous=True, **_config), 'gs')

class _CFS:
    s3: 's3fs.S3FileSystem' = _s3_filesystem
    s3a: 's3fs.S3FileSystem' = _s3_async_filesystem

    minio: 's3fs.S3FileSystem' = _minio_filesystem
    minioa: 's3fs.S3FileSystem' = _minio_async_filesystem

    s3c: 's3fs.S3FileSystem' = _s3compat_filesystem
    s3ca: 's3fs.S3FileSystem' = _s3compat_async_filesystem

    gs: 'gcsfs.GCSFileSystem' = _gcs_filesystem
    gsa: 'gcsfs.GCSFileSystem' = _gcs_async_filesystem

class _AsyncSyncS3Accessor(_NormalAccessor):
    # Sync methods
    info = _CFS.s3.info
    stat = _CFS.s3.stat
    open = _CFS.s3.open
    listdir = _CFS.s3.ls

    mkdir = _CFS.s3.mkdir
    unlink = _CFS.s3.rm_file
    rmdir = _CFS.s3.rmdir

    rename = _CFS.s3.rename
    replace = _CFS.s3.rename
    remove = _CFS.s3.rm

    # Async Methods
    async_info = _CFS.s3a.async_info
    async_stat = func_as_method_coro(_CFS.s3.stat)
    async_listdir = _CFS.s3a.async_list_objects
    async_open = _CFS.s3a._open

    async_mkdir = _CFS.s3a.async_mkdir
    async_unlink = _CFS.s3a.async_rm_file
    async_rmdir = _CFS.s3a.async_rmdir

    async_rename = func_as_method_coro(_CFS.s3.rename)
    async_replace = func_as_method_coro(_CFS.s3.rename)
    async_remove = _CFS.s3a.async_rm
    async_touch = func_as_method_coro(_CFS.s3.touch)



class _AsyncSyncGSAccessor(_NormalAccessor):
    # Sync methods
    stat = _CFS.gs.stat
    open = _CFS.gs.open
    listdir = _CFS.gs.ls

    mkdir = _CFS.gs.mkdir
    unlink = _CFS.gs.rm_file
    rmdir = _CFS.gs.rmdir

    rename = _CFS.gs.rename
    replace = _CFS.gs.rename
    remove = _CFS.gs.rm
    touch = _CFS.gs.touch

    # Async Methods
    async_stat = func_as_method_coro(_CFS.gs.stat)
    async_listdir = _CFS.gsa.async_list_objects

    async_mkdir = _CFS.gsa.async_mkdir
    async_unlink = _CFS.gsa.async_rm_file
    async_rmdir = _CFS.gsa.async_rmdir

    async_rename = func_as_method_coro(_CFS.gs.rename)
    async_replace = func_as_method_coro(_CFS.gs.rename)
    async_remove = _CFS.gsa.async_rm
    async_touch = func_as_method_coro(_CFS.gs.touch)


_async_sync_s3_accessor = _AsyncSyncS3Accessor()
_async_sync_gs_accessor = _AsyncSyncGSAccessor()


class AsyncSyncS3PurePath(PurePath):
    def _init(self, template: Optional[PurePath] = None):
        self._accessor = _async_sync_s3_accessor

    def __new__(cls, *args):
        if cls is AsyncSyncS3PurePath: cls = PureAsyncSyncS3WindowsPath if os.name == 'nt' else PureAsyncSyncS3PosixPath
        return cls._from_parts(args)

    def _new(self, *parts):
        """Create a new `Path` child of same type."""
        return type(self)(*parts)


class PureAsyncSyncS3PosixPath(AsyncSyncS3PurePath):
    """PurePath subclass for non-Windows systems.
    On a POSIX system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _async_sync_posix_flavour
    _pathlike = posixpath
    __slots__ = ()



class PureAsyncSyncS3WindowsPath(AsyncSyncS3PurePath):
    """PurePath subclass for Windows systems.
    On a Windows system, instantiating a PurePath should return this object.
    However, you can also instantiate it directly on any system.
    """
    _flavour = _async_sync_windows_flavour
    _pathlike = ntpath
    __slots__ = ()


class AsyncSyncS3Path(Path, AsyncSyncS3PurePath):
    """
    Our customized class that incorporates both sync and async methods
    """
    _flavour = _async_sync_windows_flavour if os.name == 'nt' else _async_sync_posix_flavour
    _accessor = _async_sync_s3_accessor
    _pathlike = posixpath
    _prefix = 's3'

    def _init(self, template: Optional[AsyncSyncS3Path] = None):
        self._accessor = _async_sync_s3_accessor
        self._closed = False
        self._fileio = None

    def __new__(cls, *parts, **kwargs):
        if cls is AsyncSyncS3Path: cls = AsyncSyncS3WindowsPath if os.name == 'nt' else AsyncSyncS3PosixPath
        self = cls._from_parts(parts, init=False)
        if not self._flavour.is_supported:
            name: str = cls.__name__
            raise NotImplementedError(f"cannot instantiate {name} on your system")

        self._init()
        return self

    @property
    def _path(self) -> str:
        return str(self)
    
    @property
    def _cloudpath(self) -> str:
        """
        Returns the `__fspath__` string representation without the uri_scheme
        """
        if self._prefix in self.parts[0]: return self._pathlike.join(*self.parts[1:])
        return self._pathlike.join(*self.parts)
    
    @property
    def _cloudstr(self) -> str:
        """
        Reconstructs the proper cloud URI
        """
        if self._prefix not in self.parts[0]: return self._prefix + '://' + '/'.join(self.parts)
        return self._prefix + '://' + '/'.join(self.parts[1:])
    
    @property
    def string(self) -> str:
        return str(self)
    
    @property
    def filename_(self) -> str:
        """
        Returns the filename if is file, else ''
        """
        if self.is_file(): return self.parts[-1]
        return ''

    @property
    def ext_(self) -> str:
        """
        Returns the extension for a file
        """
        return self.suffix

    @property
    def extension(self) -> str:
        """
        Returns the extension for a file
        """
        return self.suffix
    
    @property
    def is_cloud(self) -> bool:
        return True
    
    @property
    def exists_(self) -> bool:
        return self.exists()
    
    @property
    def is_file_(self) -> bool:
        return self.is_file()
    
    @property
    def is_dir_(self) -> bool:
        return self.is_dir()
    
    @property
    def home_(self) -> AsyncSyncS3Path:
        return self.home()

    @property
    async def async_exists_(self) -> bool:
        return await self.async_exists()
    
    @property
    async def async_is_file_(self) -> bool:
        return await self.async_is_file()
    
    @property
    async def async_is_dir_(self) -> bool:
        return await self.async_is_dir()
    
    @property
    async def async_home_(self) -> AsyncSyncS3Path:
        return await self.async_home()
    
    
    def open(self, mode: FileMode = 'r', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IO[Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return self._accessor.open(self._cloudpath, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline)

    
    def async_open(self, mode: FileMode = 'r', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return get_handle(self._accessor.open(self._cloudpath, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline))
    
    def reader(self, mode: FileMode = 'r', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IO[Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return self._accessor.open(self._cloudpath, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline)
    
    def async_reader(self, mode: FileMode = 'r', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return open_file(self._accessor.open(self._cloudpath, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline))
    
    def appender(self, mode: FileMode = 'a', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IO[Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return self._accessor.open(self._cloudpath, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline)
    
    def async_appender(self, mode: FileMode = 'a', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return open_file(self._accessor.open(self._cloudpath, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline))
    
    def writer(self, mode: FileMode = 'w', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IO[Union[str, bytes]]:
        """
        Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return self._accessor.open(self._cloudpath, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline)
    
    def async_writer(self, mode: FileMode = 'w', buffering: int = -1, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE, **kwargs) -> IterableAIOFile:
        """
        Asyncronously Open the file pointed by this path and return a file object, as
        the built-in open() function does.
        """
        return open_file(self._accessor.open(self._cloudpath, mode=mode, buffering=buffering, encoding=encoding, errors=errors, newline=newline))

    def read_text(self, encoding: str | None = DEFAULT_ENCODING, errors: str | None = ON_ERRORS) -> str:
        with self.open('r', encoding=encoding, errors=errors) as file:
            return file.read()

    async def async_read_text(self, encoding: str | None = DEFAULT_ENCODING, errors: str | None = ON_ERRORS) -> str:
        async with self.async_open('r', encoding=encoding, errors=errors) as file:
            return await file.read()

    def read_bytes(self) -> bytes:
        with self.open('rb') as file:
            return file.read()

    async def async_read_bytes(self) -> bytes:
        async with self.async_open('rb') as file:
            return await file.read()

    def write_bytes(self, data: bytes) -> int:
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        view = memoryview(data)
        with self.open(mode='wb') as f:
            return f.write(data)

    async def async_write_bytes(self, data: bytes) -> int:
        """
        Open the file in bytes mode, write to it, and close the file.
        """
        # type-check for the buffer interface before truncating the file
        view = memoryview(data)
        async with self.async_open(mode='wb') as f:
            return await f.write(data)

    def append_text(self, data: str, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        with self.open(mode='a', encoding=encoding, errors=errors, newline=newline) as f:
            n = f.write(data)
            n += f.write(newline)
            return n

    async def async_append_text(self, data: str, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        async with self.async_open(mode='a', encoding=encoding, errors=errors, newline=newline) as f:
            n = await f.write(data)
            n += await f.write(newline)
            return n

    def write_text(self, data: str, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        with self.open(mode='w', encoding=encoding, errors=errors, newline=newline) as f:
            return f.write(data)

    async def async_write_text(self, data: str, encoding: Optional[str] = DEFAULT_ENCODING, errors: Optional[str] = ON_ERRORS, newline: Optional[str] = NEWLINE) -> int:
        """
        Open the file in text mode, write to it, and close the file.
        """
        if not isinstance(data, str): raise TypeError(f'data must be str, not {type(data).__name__}')
        async with self.async_open(mode='w', encoding=encoding, errors=errors, newline=newline) as f:
            return await f.write(data)


    def touch(self, truncate: bool = True, data = None, exist_ok: bool = True, **kwargs):
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        if exist_ok:
            try: self._accessor.stat(self._cloudpath)
            # Avoid exception chaining
            except OSError: pass
            else: return
        self._accessor.touch(self._cloudpath, truncate = truncate, data = data, **kwargs)

    async def async_touch(self, truncate: bool = True, data = None, exist_ok: bool = True, **kwargs):
        """
        Create this file with the given access mode, if it doesn't exist.
        """
        if exist_ok:
            try: await self._accessor.async_stat(self._cloudpath)
            # Avoid exception chaining
            except OSError: pass
            else: return
        await self._accessor.async_touch(self._cloudpath, truncate = truncate, data = data, **kwargs)



class AsyncSyncS3PosixPath(PosixPath, AsyncSyncS3Path, PureAsyncSyncS3PosixPath):
    __slots__ = ()


class AsyncSyncS3WindowsPath(WindowsPath, AsyncSyncS3Path, PureAsyncSyncS3WindowsPath):
    __slots__ = ()

    def is_mount(self) -> int:
        raise NotImplementedError("AsyncSyncPath.is_mount() is unsupported on this system")

    async def async_is_mount(self) -> int:
        raise NotImplementedError("AsyncSyncPath.async_is_mount() is unsupported on this system")



__all__ = (
    "classproperty",
    "classproperty_v2",
    "ClasspropertyMeta",
    "cached_property"
)


from typing import Optional, Any

class _classproperty(property):
    """
    The use this class as a decorator for your class property.
    Example:
        @classproperty
        def prop(cls):
            return "value"
    """

    def __get__(self, *args, **_) -> object:
        """
        This method gets called when a property value is requested.
        @return: The value of the property.
        """
        # apply the __get__ on the class, which is the second argument
        return super(_classproperty, self).__get__(args[1])

    def __set__(self, cls_or_instance, value: object) -> None:
        """
        This method gets called when a property value should be set.
        @param cls_or_instance: The class or instance of which the property should be changed.
        @param value: The new value.
        """
        # call this method only on the class, not the instance
        super(_classproperty, self).__set__(self.__get_class(cls_or_instance), value)

    def __delete__(self, cls_or_instance) -> None:
        """
        This method gets called when a property should be deleted.
        @param cls_or_instance: The class or instance of which the property should be deleted.
        """
        # call this method only on the class, not the instance
        super(_classproperty, self).__delete__(self.__get_class(cls_or_instance))

    def __get_class(self, cls_or_instance) -> type:
        """
        Get the class of an object if one is provided.
        @param cls_or_instance: Either an object or a class.
        @return: The class of the object or just the class again.
        """
        if isinstance(cls_or_instance, type): return cls_or_instance
        return type(cls_or_instance)

class _ClasspropertyMeta(type):
    """
    The class that uses the classproperty decorator must use this meta class if the setting and deleting of a class
    property should be supported.
    """
    def __setattr__(self, name: str, value: object) -> None:
        """
        Override of __setattr__ method to allow a classproperty.setter.
        @param name: The name of the attribute that should get a new value.
        @param value: The new value of the attribute.
        """
        cp_obj: Optional[_classproperty] = self.__get_classproperty_attr(name)
        if cp_obj: cp_obj.__set__(self, value)
        else: super(_ClasspropertyMeta, self).__setattr__(name, value)

    def __delattr__(self, name: str):
        """
        Override of __delattr__ method to allow a classproperty.deleter.
        @param str: The name of the attribute to delete.
        """
        cp_obj: Optional[_classproperty] = self.__get_classproperty_attr(name)
        if cp_obj: cp_obj.__delete__(self)
        else: super(_ClasspropertyMeta, self).__delattr__(name)

    def __get_classproperty_attr(self, name: str) -> Optional[_classproperty]:
        """
        Get a classproperty attribute from this class with the given name.
        @param name: The name of the attribute.
        @return: The classproperty object for the attribute. Or 'None' if it wasn't found.
        """
        # iterate through MRO list to get all attributes
        if (name in self.__dict__ and isinstance(self.__dict__[name], _classproperty)): return self.__dict__[name]
        else: return None

class _cached_property:
    """
    Decorator that converts a method with a single self argument into a
    property cached on the instance.

    A cached property can be made out of an existing method:
    (e.g. ``url = cached_property(get_absolute_url)``).
    source: django
    """
    name = None

    @staticmethod
    def func(instance):
        raise TypeError(
            'Cannot use cached_property instance without calling '
            '__set_name__() on it.'
        )

    def __init__(self, func):
        self.real_func = func
        self.__doc__ = getattr(func, '__doc__')

    def __set_name__(self, owner, name):
        if self.name is None:
            self.name = name
            self.func = self.real_func
        elif name != self.name:
            raise TypeError(
                "Cannot assign the same cached_property to two different names "
                "(%r and %r)." % (self.name, name)
            )

    def __get__(self, instance, cls=None):
        """
        Call the function and put the return value in instance.__dict__ so that
        subsequent attribute access on the instance returns the cached value
        instead of calling cached_property.__get__().
        """
        if instance is None:
            return self
        res = instance.__dict__[self.name] = self.func(instance)
        return res


class cached_property(property):
    """
    Descriptor that mimics @property but caches output in member variable.

    From tensorflow_datasets

    Built-in in functools from Python 3.8.
    """

    def __get__(self, obj, objtype=None):
        # See docs.python.org/3/howto/descriptor.html#properties
        if obj is None:
            return self
        if self.fget is None:
            raise AttributeError("unreadable attribute")
        attr = "__cached_" + self.fget.__name__
        cached = getattr(obj, attr, None)
        if cached is None:
            cached = self.fget(obj)
            setattr(obj, attr, cached)
        return cached


class classproperty_v2:
    """
    Decorator that converts a method with a single cls argument into a property
    that can be accessed directly from the class.
    Simpler than original. 
    Source: django 
    """
    def __init__(self, method=None):
        self.fget = method

    def __get__(self, instance, cls=None):
        return self.fget(cls)

    def getter(self, method):
        self.fget = method
        return self

class classproperty(_classproperty):
    pass

class ClasspropertyMeta(_ClasspropertyMeta):
    pass
